@@ -7,7 +7,8 @@ use crate::{
     checkout::Checkout,
     context::Context,
     customer::{Address, Contact},
-    error::{new_invalid_state_error, Error},
+    error::{new_application_error, new_invalid_state_error, Error},
+    internal_context::InternalContext,
     invoice::Invoice,
     item::Item,
     payment::PaymentProcessor,
@@ -21,7 +22,7 @@ pub mod handlers;
 pub struct Order {
     pub id: String,
     pub state: OrderState,
-    pub reconciling_state: bool,
+    pub state_reconciled: bool,
     pub contact: Contact,
     pub shipping_address: Address,
     pub items: Vec<Item>,
@@ -39,6 +40,12 @@ pub enum OrderState {
     Cancelled, // a terminal state
 }
 
+#[derive(Serialize, Deserialize)]
+pub enum LookupKey {
+    CheckoutId(String),
+    OrderId(String),
+}
+
 #[async_trait]
 pub trait OrderStore {
     async fn create_order(&mut self, order: &Order) -> Result<(), Error>;
@@ -72,12 +79,13 @@ pub trait OrderEvents {
 
 impl Order {
     pub async fn new_from_checkout<C: Context + Send>(
+        internal_ctx: &InternalContext,
         ctx: &mut C,
         checkout: &Checkout,
     ) -> Result<Self, Error> {
         let order = Order {
             id: Uuid::new_v4().to_string(),
-            reconciling_state: false,
+            state_reconciled: false,
             state: OrderState::PaymentInProgress,
             contact: checkout.contact.as_ref().unwrap().clone(),
             shipping_address: checkout.shipping_address.as_ref().unwrap().clone(),
@@ -88,17 +96,22 @@ impl Order {
         };
 
         ctx.create_order(&order).await?;
+
+        publish_state_change_event(internal_ctx, LookupKey::CheckoutId(checkout.id.clone()))
+            .await?;
+
         Ok(order)
     }
 
     pub async fn new_from_checkout_missing_payment<C: Context + Send>(
+        internal_ctx: &InternalContext,
         ctx: &mut C,
         checkout: &Checkout,
-        payment_args: Option<<C as PaymentProcessor>::ProcessArgs>,
+        payment_args: <C as PaymentProcessor>::ProcessArgs,
     ) -> Result<Self, Error> {
         let mut order = Order {
             id: Uuid::new_v4().to_string(),
-            reconciling_state: false,
+            state_reconciled: false,
             state: OrderState::PaymentInProgress,
             contact: checkout.contact.as_ref().unwrap().clone(),
             shipping_address: checkout.shipping_address.as_ref().unwrap().clone(),
@@ -120,51 +133,96 @@ impl Order {
         order.payment_id = Some(payment.id.clone());
 
         ctx.create_order(&order).await?;
+
+        publish_state_change_event(internal_ctx, LookupKey::CheckoutId(checkout.id.clone()))
+            .await?;
+
         Ok(order)
     }
 }
 
 impl Order {
-    fn enter_pending_confirmation_state(&mut self) {
-        self.state = OrderState::PendingConfirmation;
+    async fn assert_reconciled(&self) -> Result<(), Error> {
+        match self.state_reconciled {
+            true => Ok(()),
+            false => Err(new_application_error(
+                "ORDER_STATE_RECONCILING",
+                "the order state is reconciling. please try again in a few moments",
+            )),
+        }
     }
 
-    fn enter_confirmed_state(&mut self) {
-        self.state = OrderState::Confirmed;
+    async fn transition_state(
+        &mut self,
+        internal_ctx: &InternalContext,
+        to: OrderState,
+    ) -> Result<(), Error> {
+        self.assert_reconciled().await?;
+
+        self.state = to;
+        self.state_reconciled = false;
+
+        publish_state_change_event(internal_ctx, LookupKey::OrderId(self.id.clone())).await
     }
 
-    fn enter_cancelled_state(&mut self) {
-        self.state = OrderState::Cancelled;
-    }
-}
-
-impl Order {
-    fn ensure_pending_confirmation_state(&mut self) -> Result<(), Error> {
+    async fn ensure_pending_confirmation_state(
+        &mut self,
+        internal_ctx: &InternalContext,
+    ) -> Result<(), Error> {
         match self.state {
             OrderState::PendingConfirmation => Ok(()),
-            OrderState::PaymentInProgress => Ok(self.enter_pending_confirmation_state()),
+            OrderState::PaymentInProgress => {
+                self.transition_state(internal_ctx, OrderState::PendingConfirmation)
+                    .await
+            }
             _ => Err(new_invalid_state_error(
                 "should be in the PaymentInProgress state",
             )),
         }
     }
 
-    fn ensure_confirmed_state(&mut self) -> Result<(), Error> {
+    async fn ensure_confirmed_state(
+        &mut self,
+        internal_ctx: &InternalContext,
+    ) -> Result<(), Error> {
         match self.state {
             OrderState::Confirmed => Ok(()),
-            OrderState::PendingConfirmation => Ok(self.enter_confirmed_state()),
+            OrderState::PendingConfirmation => {
+                self.transition_state(internal_ctx, OrderState::Confirmed)
+                    .await
+            }
             _ => Err(new_invalid_state_error(
                 "should be in the PendingConfirmation state",
             )),
         }
     }
 
-    fn ensure_cancelled_state(&mut self) -> Result<(), Error> {
+    async fn ensure_fulfilled_state(
+        &mut self,
+        internal_ctx: &InternalContext,
+    ) -> Result<(), Error> {
+        match self.state {
+            OrderState::Fulfilled => Ok(()),
+            OrderState::Confirmed => {
+                self.transition_state(internal_ctx, OrderState::Fulfilled)
+                    .await
+            }
+            _ => Err(new_invalid_state_error("should be in the Confirmed state")),
+        }
+    }
+
+    async fn ensure_cancelled_state(
+        &mut self,
+        internal_ctx: &InternalContext,
+    ) -> Result<(), Error> {
         match self.state {
             OrderState::Cancelled => Ok(()),
             OrderState::PaymentInProgress
             | OrderState::PendingConfirmation
-            | OrderState::Confirmed => Ok(self.enter_cancelled_state()),
+            | OrderState::Confirmed => {
+                self.transition_state(internal_ctx, OrderState::Cancelled)
+                    .await
+            }
             _ => Err(new_invalid_state_error(
                 "should be in the PaymentInProgress, PendingConfirmation or Confirmed states",
             )),
@@ -175,19 +233,56 @@ impl Order {
 impl Order {
     pub async fn prompt_confirmation<C: Context + Send>(
         &mut self,
+        internal_ctx: &InternalContext,
         ctx: &mut C,
     ) -> Result<(), Error> {
-        self.ensure_pending_confirmation_state()?;
+        self.ensure_pending_confirmation_state(internal_ctx).await?;
         ctx.update_order(self).await
     }
 
-    pub async fn confirm<C: Context + Send>(&mut self, ctx: &mut C) -> Result<(), Error> {
-        self.ensure_confirmed_state()?;
+    pub async fn confirm<C: Context + Send>(
+        &mut self,
+        internal_ctx: &InternalContext,
+        ctx: &mut C,
+    ) -> Result<(), Error> {
+        self.ensure_confirmed_state(internal_ctx).await?;
         ctx.update_order(self).await
     }
 
-    pub async fn cancel<C: Context + Send>(&mut self, ctx: &mut C) -> Result<(), Error> {
-        self.ensure_cancelled_state()?;
+    pub async fn fulfill<C: Context + Send>(
+        &mut self,
+        internal_ctx: &InternalContext,
+        ctx: &mut C,
+    ) -> Result<(), Error> {
+        self.ensure_fulfilled_state(internal_ctx).await?;
+        ctx.update_order(self).await
+    }
+
+    pub async fn cancel<C: Context + Send>(
+        &mut self,
+        internal_ctx: &InternalContext,
+        ctx: &mut C,
+    ) -> Result<(), Error> {
+        self.ensure_cancelled_state(internal_ctx).await?;
         ctx.update_order(self).await
     }
 }
+
+async fn publish_state_change_event(
+    internal_ctx: &InternalContext,
+    lookup_key: LookupKey,
+) -> Result<(), Error> {
+    internal_ctx
+        .order_events_topic
+        .clone()
+        .publish(lookup_key)
+        .await
+        .map_err(|_| {
+            new_application_error(
+                "UPSTREAM_ERROR",
+                "couldn't send message to order events topic",
+            )
+        })?;
+
+    Ok(())
+}
