@@ -46,6 +46,18 @@ pub trait PaymentProcessor {
     type PrivateData: Sync;
     type PublicData: Sync;
 
+    async fn initiate_payment(
+        &mut self,
+        _payment: &Payment<Self::PrivateData, Self::PublicData, Self::ProcessArgs>,
+        _args: &Self::ProcessArgs,
+    ) -> Result<PaymentData<Self::PrivateData, Self::PublicData>, Error> {
+        Ok(PaymentData {
+            correlation_id: None,
+            private_data: None,
+            public_data: None,
+        })
+    }
+
     async fn process_payment(
         &mut self,
         payment: &Payment<Self::PrivateData, Self::PublicData, Self::ProcessArgs>,
@@ -153,7 +165,7 @@ async fn new<'a, C: Context + Send>(
     ctx: &mut C,
     initiator: Initiator<'a>,
     charge_amount: Money,
-    maybe_args: Option<<C as PaymentProcessor>::ProcessArgs>,
+    args: <C as PaymentProcessor>::ProcessArgs,
 ) -> Result<
     Payment<
         <C as PaymentProcessor>::PrivateData,
@@ -177,34 +189,9 @@ async fn new<'a, C: Context + Send>(
         },
     };
 
+    payment.data = ctx.initiate_payment(&payment, &args).await?;
     ctx.create_payment(&payment).await?;
 
-    if let Some(args) = maybe_args {
-        let result = ctx.process_payment(&mut payment, &args).await?;
-        payment.data = result.data;
-        match result.state {
-            InternalPaymentState::Incomplete | InternalPaymentState::Failed => {
-                payment.state = PaymentState::Idle;
-            }
-            InternalPaymentState::Succeeded => payment.state = PaymentState::Succeeded,
-            InternalPaymentState::Cancelled => payment.state = PaymentState::Cancelled,
-        }
-
-        ctx.update_payment(&payment).await?;
-
-        match payment.state {
-            PaymentState::Succeeded => match initiator {
-                Initiator::Checkout(co) => co.complete(ctx).await?,
-                Initiator::Order(od) => od.prompt_confirmation(ctx).await?,
-            },
-            PaymentState::Cancelled => match initiator {
-                Initiator::Checkout(co) => co.revert(ctx).await?,
-                Initiator::Order(od) => od.cancel(ctx).await?,
-            },
-            _ => {}
-        }
-    }
-
     Ok(payment)
 }
 
@@ -221,6 +208,9 @@ pub async fn queue_for_processing<C: Context + Send>(
 ) -> Result<(), Error> {
     match &payment.state {
         PaymentState::Idle => {
+            payment.state = PaymentState::Processing(args);
+            ctx.update_payment(payment).await?;
+
             internal_ctx
                 .payment_processing_topic
                 .clone()
@@ -232,8 +222,8 @@ pub async fn queue_for_processing<C: Context + Send>(
                         "couldn't send message to payment processing topic",
                     )
                 })?;
-            payment.state = PaymentState::Processing(args);
-            ctx.update_payment(payment).await
+
+            Ok(())
         }
         PaymentState::Processing(_) => Err(new_application_error(
             "PAYMENT_PROCESSING",
@@ -247,6 +237,7 @@ pub async fn queue_for_processing<C: Context + Send>(
 }
 
 pub async fn process<'a, C: Context + Send>(
+    internal_ctx: &InternalContext,
     ctx: &mut C,
     payment: &mut Payment<
         <C as PaymentProcessor>::PrivateData,
@@ -271,12 +262,12 @@ pub async fn process<'a, C: Context + Send>(
 
             match payment.state {
                 PaymentState::Succeeded => match initiator {
-                    Initiator::Checkout(co) => co.complete(ctx).await,
-                    Initiator::Order(od) => od.prompt_confirmation(ctx).await,
+                    Initiator::Checkout(co) => co.complete(internal_ctx, ctx).await,
+                    Initiator::Order(od) => od.prompt_confirmation(internal_ctx, ctx).await,
                 },
                 PaymentState::Cancelled => match initiator {
                     Initiator::Checkout(co) => co.revert(ctx).await,
-                    Initiator::Order(od) => od.cancel(ctx).await,
+                    Initiator::Order(od) => od.cancel(internal_ctx, ctx).await,
                 },
                 _ => Ok(()),
             }
@@ -286,6 +277,7 @@ pub async fn process<'a, C: Context + Send>(
 }
 
 pub async fn cancel<'a, C: Context + Send>(
+    internal_ctx: &InternalContext,
     ctx: &mut C,
     payment: &mut Payment<
         <C as PaymentProcessor>::PrivateData,
@@ -310,7 +302,7 @@ pub async fn cancel<'a, C: Context + Send>(
 
             match initiator {
                 Initiator::Checkout(co) => co.revert(ctx).await,
-                Initiator::Order(od) => od.cancel(ctx).await,
+                Initiator::Order(od) => od.cancel(internal_ctx, ctx).await,
             }
         }
         PaymentState::Processing(_) => Err(new_application_error(
