@@ -14,8 +14,10 @@ pub struct ItemPrice {
 pub trait PriceCalculator {
     async fn calculate_item_prices(
         &mut self,
-        currency: &Currency,
-        promo_codes: &Vec<String>,
-        items: &Vec<Item>,
-    ) -> Result<Vec<ItemPrice>, Error>;
+        _currency: &Currency,
+        _promo_codes: &Vec<String>,
+        _items: &Vec<Item>,
+    ) -> Result<Vec<ItemPrice>, Error> {
+        Ok(vec![])
+    }
 }
